# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
import re
from logging import Logger
from pathlib import Path
from typing import Dict, List, Optional, Sequence

import libcst
from libcst.codemod import CodemodContext
from libcst.codemod.visitors import AddImportsVisitor
from pyre_extensions import override

from ..commands.command import CommandArguments, ErrorSuppressingCommand
from ..configuration import Configuration
from ..errors import PathsToErrors, PyreError, UserError
from ..filesystem import path_exists
from ..repository import Repository

LOG: Logger = logging.getLogger(__name__)


MISSING_ATTRIBUTE_ANNOTATION_ERROR_CODE = 4
UNBOUND_NAME_ERROR_CODE = 10


def _is_sqlalchemy_error(error: PyreError) -> bool:
    return bool(
        re.search(
            r"has type .*?Column\[.*but no type is specified.", error["description"]
        )
    )


class SupportSqlalchemy(ErrorSuppressingCommand):
    def __init__(
        self,
        command_arguments: CommandArguments,
        *,
        local_root: Path,
        paths: Sequence[Path],
        repository: Repository,
    ) -> None:
        super().__init__(command_arguments, repository=repository)
        self._local_root = local_root
        self._paths: Optional[Sequence[Path]] = paths if len(paths) > 0 else None

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "SupportSqlalchemy":
        command_arguments = CommandArguments.from_arguments(arguments)
        return SupportSqlalchemy(
            command_arguments=command_arguments,
            local_root=arguments.local_root,
            paths=arguments.paths,
            repository=repository,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(SupportSqlalchemy, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "-l",
            "--local-root",
            help="Path to directory with local configuration",
            type=path_exists,
            required=True,
        )
        parser.add_argument(
            "paths", help="Paths using sqlalchemy", type=path_exists, nargs="*"
        )

    @staticmethod
    def _get_sqlalchemy_errors(
        path_wise_errors: Dict[str, List[PyreError]],
        filter_paths: Optional[Sequence[Path]],
    ) -> PathsToErrors:
        all_pathwise_sqlalchemy_errors = {
            Path(pathname): [error for error in errors if _is_sqlalchemy_error(error)]
            for pathname, errors in path_wise_errors.items()
        }

        nonempty_pathwise_sqlalchemy_errors = {
            path: errors
            for path, errors in all_pathwise_sqlalchemy_errors.items()
            if len(errors) > 0
        }

        if filter_paths is None:
            return nonempty_pathwise_sqlalchemy_errors

        return {
            path: errors
            for path, errors in nonempty_pathwise_sqlalchemy_errors.items()
            if path in filter_paths
        }

    def _annotate_sqlalchemy_files(
        self, configuration: Configuration, sqlalchemy_path_wise_errors: PathsToErrors
    ) -> None:
        paths = [str(path) for path in sqlalchemy_path_wise_errors.keys()]
        pyre_output = configuration.run_pyre(
            arguments=[
                "--strict",
                "-l",
                str(self._local_root),
                "--noninteractive",
                "infer-v2",
                "--in-place",
                "--dequalify",
                *paths,
            ],
            description="Running `pyre infer-v2`",
            should_clean=self._should_clean,
            stderr_flag=None,
            command_input=None,
        )
        if pyre_output is None:
            raise UserError("Couldn't annotate sqlalchemy files.")

    def _import_annotations_from_future(
        self, sqlalchemy_path_wise_errors: PathsToErrors
    ) -> None:
        """We need this because the original sqlalchemy types aren't generic
        and will fail at runtime."""
        LOG.info("Importing necessary annotations...")
        context = CodemodContext()
        AddImportsVisitor.add_needed_import(context, "__future__", "annotations")
        paths = list(sqlalchemy_path_wise_errors.keys())
        for path in paths:
            source = libcst.parse_module(path.read_text())
            modified_tree = AddImportsVisitor(context).transform_module(source)
            path.write_text(modified_tree.code)

    @override
    def run(self) -> None:
        local_configuration_path = self._local_root / ".pyre_configuration.local"
        local_configuration = Configuration(local_configuration_path)

        unannotated_attribute_errors = local_configuration.get_errors(
            only_fix_error_code=MISSING_ATTRIBUTE_ANNOTATION_ERROR_CODE,
            strict=True,
            should_clean=self._should_clean,
        )

        sqlalchemy_path_wise_errors = self._get_sqlalchemy_errors(
            unannotated_attribute_errors.paths_to_errors, self._paths
        )

        if len(sqlalchemy_path_wise_errors) == 0:
            LOG.warning("No paths with missing annotations. Exiting...")
            return

        LOG.info("Found errors: %s", sqlalchemy_path_wise_errors)

        LOG.info(
            "Annotating the following sqlalchemy files: `%s`",
            list(sqlalchemy_path_wise_errors.keys()),
        )
        self._annotate_sqlalchemy_files(
            local_configuration, sqlalchemy_path_wise_errors
        )
        self._import_annotations_from_future(sqlalchemy_path_wise_errors)

        self._get_and_suppress_errors(local_configuration)

        title = "Suppress errors for {}".format(self._local_root)
        self._repository.commit_changes(commit=(not self._no_commit), title=title)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
The global strictness command sets the global configuration's strictness field and adjusts the local configurations to preserve semantics.

For example, if a global configuration is strict and local configurations are non-strict, running the command with strict=False will set the global configuration to
strict=False, remove strict=False, and set any unset local configurations to strict=True. Alternatively, running with strict=True will set the global configuration to
strict, remove any configurations strict=True, and any unset configurations will be set to False.
"""

from __future__ import annotations

import argparse
import logging

from ..configuration import Configuration
from ..repository import Repository
from .command import Command

LOG: logging.Logger = logging.getLogger(__name__)


class GlobalStrictness(Command):
    def __init__(self, *, repository: Repository, strict: bool) -> None:
        self.repository = repository
        self.strict = strict

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> GlobalStrictness:
        return GlobalStrictness(repository=repository, strict=arguments.strict)

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(GlobalStrictness, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "--strict",
            type=bool,
            help="Whether to make the global configuration strict or non-strict",
        )

    def run(self) -> None:
        global_configuration = Configuration(Configuration.find_project_configuration())

        LOG.info(f"Setting global strictness to {self.strict}...")
        global_configuration.strict = self.strict
        global_configuration.write()
        configurations = Configuration.gather_local_configurations()

        LOG.info(
            f"Found {len(configurations)} configurations. Setting unset configurations to strict={not self.strict}..."
        )
        additions = 0
        removals = 0
        for configuration in configurations:
            if configuration.strict is None:
                additions += 1
                configuration.strict = not self.strict
                configuration.write()
            if configuration.strict == self.strict:
                removals += 1
                configuration.strict = None
                configuration.write()
        LOG.info(
            f"Done setting {additions} local configurations to strict={not self.strict} and removing {removals} unnecessary settings."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module specifies how one can tune the backend's map-reduce operations using
a "scheduler policies" configuration file. This can be used to improve
performance on specific machines.
"""

import dataclasses
import json
from pathlib import Path
from typing import Dict, Optional, Union

from .exceptions import InvalidConfiguration


@dataclasses.dataclass(frozen=True)
class FixedChunkSize:
    minimum_chunks_per_worker: int
    preferred_chunk_size: int
    minimum_chunk_size: Optional[int] = None


@dataclasses.dataclass(frozen=True)
class FixedChunkCount:
    minimum_chunk_size: int
    preferred_chunks_per_worker: int
    minimum_chunks_per_worker: Optional[int] = None


def optional_positive_int_member(
    object: Dict[str, object], key: str, identifier: str
) -> Optional[int]:
    if key not in object:
        return None

    value = object[key]
    if not isinstance(value, int):
        raise InvalidConfiguration(
            f"Invalid scheduler policy for `{identifier}`: expected integer for `{key}`, got `{value}`"
        )

    if value < 1:
        raise InvalidConfiguration(
            f"Invalid scheduler policy for `{identifier}`: expected positive integer for `{key}`, got `{value}`"
        )

    return value


def positive_int_member(object: Dict[str, object], key: str, identifier: str) -> int:
    value = optional_positive_int_member(object, key, identifier)

    if value is None:
        raise InvalidConfiguration(
            f"Invalid scheduler policy for `{identifier}`: missing `{key}` key in `{object}`"
        )

    return value


@dataclasses.dataclass(frozen=True)
class SchedulerPolicy:
    value: Union[FixedChunkSize, FixedChunkCount]

    @staticmethod
    def from_json(value: object, identifier: str) -> "SchedulerPolicy":
        if not isinstance(value, dict):
            raise InvalidConfiguration(
                f"Invalid scheduler policy for `{identifier}`: expected object, but got `{type(value).__name__}`"
            )

        if "kind" not in value:
            raise InvalidConfiguration(
                f"Invalid scheduler policy for `{identifier}`: missing `kind` key in `{value}`"
            )

        kind = value["kind"]
        if kind == "fixed_chunk_size":
            minimum_chunk_size = optional_positive_int_member(
                value, "minimum_chunk_size", identifier
            )
            minimum_chunks_per_worker = positive_int_member(
                value, "minimum_chunks_per_worker", identifier
            )
            preferred_chunk_size = positive_int_member(
                value, "preferred_chunk_size", identifier
            )
            return SchedulerPolicy(
                value=FixedChunkSize(
                    minimum_chunk_size=minimum_chunk_size,
                    minimum_chunks_per_worker=minimum_chunks_per_worker,
                    preferred_chunk_size=preferred_chunk_size,
                )
            )
        elif kind == "fixed_chunk_count":
            minimum_chunks_per_worker = optional_positive_int_member(
                value, "minimum_chunks_per_worker", identifier
            )
            minimum_chunk_size = positive_int_member(
                value, "minimum_chunk_size", identifier
            )
            preferred_chunks_per_worker = positive_int_member(
                value, "preferred_chunks_per_worker", identifier
            )
            return SchedulerPolicy(
                value=FixedChunkCount(
                    minimum_chunks_per_worker=minimum_chunks_per_worker,
                    minimum_chunk_size=minimum_chunk_size,
                    preferred_chunks_per_worker=preferred_chunks_per_worker,
                )
            )
        else:
            raise InvalidConfiguration(
                f"Invalid scheduler policy kind: got `{kind}`, expected `fixed_chunk_size` or `fixed_chunk_count`"
            )

    def to_json(self) -> Dict[str, Union[int, str]]:
        value = self.value
        if isinstance(value, FixedChunkSize):
            minimum_chunk_size = value.minimum_chunk_size
            return {
                "kind": "fixed_chunk_size",
                **(
                    {"minimum_chunk_size": minimum_chunk_size}
                    if minimum_chunk_size is not None
                    else {}
                ),
                "minimum_chunks_per_worker": value.minimum_chunks_per_worker,
                "preferred_chunk_size": value.preferred_chunk_size,
            }
        elif isinstance(value, FixedChunkCount):
            minimum_chunks_per_worker = value.minimum_chunks_per_worker
            return {
                "kind": "fixed_chunk_count",
                **(
                    {"minimum_chunks_per_worker": minimum_chunks_per_worker}
                    if minimum_chunks_per_worker is not None
                    else {}
                ),
                "minimum_chunk_size": value.minimum_chunk_size,
                "preferred_chunks_per_worker": value.preferred_chunks_per_worker,
            }
        else:
            raise AssertionError("unexpected policy")


@dataclasses.dataclass(frozen=True)
class SchedulerPolicies:
    policies: Dict[str, SchedulerPolicy] = dataclasses.field(default_factory=dict)

    @staticmethod
    def from_json(value: object) -> "SchedulerPolicies":
        if not isinstance(value, dict):
            raise InvalidConfiguration(
                f"Expected object for scheduler policies, but got `{type(value).__name__}`"
            )

        return SchedulerPolicies(
            policies={
                name: SchedulerPolicy.from_json(policy, name)
                for name, policy in value.items()
            }
        )

    @staticmethod
    def from_path(path: Path) -> "SchedulerPolicies":
        with open(path, "r") as f:
            try:
                return SchedulerPolicies.from_json(json.load(f))
            except json.JSONDecodeError as error:
                raise InvalidConfiguration(
                    f"Error while parsing `{str(path)}`: {error.lineno}:{error.colno}: {error.msg}"
                ) from error

    def to_json(self) -> Dict[str, Dict[str, Union[int, str]]]:
        return {name: policy.to_json() for name, policy in self.policies.items()}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides async stream readers and writers, in both bytes and
string formats. We use these to manage the standard input and output
connections in language servers.
"""

import abc
import asyncio
import contextlib
import io
import logging
import socket
import sys
from pathlib import Path
from typing import AsyncIterator, BinaryIO, Iterator, List, Optional, TextIO, Tuple

LOG: logging.Logger = logging.getLogger(__name__)


class ConnectionFailure(Exception):
    pass


@contextlib.contextmanager
def _connect_bytes(
    socket_path: Path,
) -> Iterator[Tuple[BinaryIO, BinaryIO]]:
    """
    Connect to the socket at given path. Once connected, create an input and
    an output stream from the socket. Both the input stream and the output
    stream are in raw binary mode: read/write APIs of the streams need to use
    `bytes` rather than `str`. The API is intended to be used like this:

    ```
    with connect(socket_path) as (input_stream, output_stream):
        # Read from input_stream and write into output_stream here
        ...
    ```

    Socket creation, connection, and closure will be automatically handled
    inside this context manager. If any of the socket operations fail, raise
    `ConnectionFailure`.
    """
    try:
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as client_socket:
            client_socket.connect(str(socket_path))
            with client_socket.makefile(
                mode="rb"
            ) as input_channel, client_socket.makefile(mode="wb") as output_channel:
                yield (input_channel, output_channel)
    except OSError as error:
        raise ConnectionFailure() from error


@contextlib.contextmanager
def connect(
    socket_path: Path,
) -> Iterator[Tuple[TextIO, TextIO]]:
    """
    This is a line-oriented higher-level API than `connect`. It can be used
    when the caller does not want to deal with the complexity of binary I/O.

    The behavior is the same as `connect`, except the streams that are created
    operates in text mode. Read/write APIs of the streams uses UTF-8 encoded
    `str` instead of `bytes`. Those operations are also line-buffered, meaning
    that the streams will automatically be flushed once the newline character
    is encountered.
    """
    with _connect_bytes(socket_path) as (input_channel, output_channel):
        yield (
            io.TextIOWrapper(
                input_channel,
                line_buffering=True,
                errors="replace",
            ),
            io.TextIOWrapper(
                output_channel,
                line_buffering=True,
                errors="replace",
            ),
        )


class AsyncBytesReader(abc.ABC):
    """
    This class defines the basic interface for async I/O input channel.
    """

    @abc.abstractmethod
    async def read_until(self, separator: bytes = b"\n") -> bytes:
        """
        Read data from the stream until `separator` is found.
        If EOF is reached before the complete separator is found, raise
        `asyncio.IncompleteReadError`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def read_exactly(self, count: int) -> bytes:
        """
        Read exactly `count` bytes.
        If EOF is reached before the complete separator is found, raise
        `asyncio.IncompleteReadError`.
        """
        raise NotImplementedError()

    async def readline(self) -> bytes:
        """
        Read one line, where "line" is a sequence of bytes ending with '\n'.
        If EOF is received and '\n' was not found, the method returns partially
        read data.
        """
        try:
            return await self.read_until(b"\n")
        except asyncio.IncompleteReadError as error:
            return error.partial


class AsyncBytesWriter(abc.ABC):
    """
    This class defines the basic interface for async I/O output channel.
    """

    @abc.abstractmethod
    async def write(self, data: bytes) -> None:
        """
        The method attempts to write the data to the underlying channel and
        flushes immediately.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def close(self) -> None:
        """
        The method closes the underlying channel and wait until the channel is
        fully closed.
        """
        raise NotImplementedError()


class AsyncTextReader:
    """
    An adapter for `AsyncBytesReader` that decodes everything it reads immediately
    from bytes to string. In other words, it tries to expose the same interfaces
    as `AsyncBytesReader` except it operates on strings rather than bytestrings.
    """

    bytes_reader: AsyncBytesReader
    encoding: str
    errors: str

    def __init__(
        self,
        bytes_reader: AsyncBytesReader,
        encoding: str = "utf-8",
        errors: str = "replace",
    ) -> None:
        self.bytes_reader = bytes_reader
        self.encoding = encoding
        self.errors = errors

    async def read_until(self, separator: str = "\n") -> str:
        separator_bytes = separator.encode(self.encoding)
        result_bytes = await self.bytes_reader.read_until(separator_bytes)
        return result_bytes.decode(self.encoding, errors=self.errors)

    async def read_exactly(self, count: int) -> str:
        result_bytes = await self.bytes_reader.read_exactly(count)
        return result_bytes.decode(self.encoding, errors=self.errors)

    async def readline(self) -> str:
        result_bytes = await self.bytes_reader.readline()
        return result_bytes.decode(self.encoding, errors=self.errors)


class AsyncTextWriter:
    """
    An adapter for `AsyncBytesWriter` that encodes everything it writes immediately
    from string to bytes. In other words, it tries to expose the same interfaces
    as `AsyncBytesWriter` except it operates on strings rather than bytestrings.
    """

    bytes_writer: AsyncBytesWriter
    encoding: str

    def __init__(self, bytes_writer: AsyncBytesWriter, encoding: str = "utf-8") -> None:
        self.bytes_writer = bytes_writer
        self.encoding = encoding

    async def write(self, data: str) -> None:
        data_bytes = data.encode(self.encoding)
        await self.bytes_writer.write(data_bytes)


class MemoryBytesReader(AsyncBytesReader):
    """
    An implementation of `AsyncBytesReader` based on a given in-memory byte string.
    """

    _data: bytes
    _cursor: int

    def __init__(self, data: bytes) -> None:
        self._data = data
        self._cursor = 0

    async def read_until(self, separator: bytes = b"\n") -> bytes:
        result = bytearray()
        start_index = self._cursor
        end_index = len(self._data)
        for index in range(start_index, end_index):
            byte = self._data[index]
            result.append(byte)
            if result.endswith(separator):
                self._cursor = index + 1
                return bytes(result)

        self._cursor = end_index
        raise asyncio.IncompleteReadError(bytes(result), None)

    async def read_exactly(self, count: int) -> bytes:
        old_cursor = self._cursor
        new_cursor = self._cursor + count
        data_size = len(self._data)
        if new_cursor <= data_size:
            self._cursor = new_cursor
            return self._data[old_cursor:new_cursor]
        else:
            self._cursor = data_size
            raise asyncio.IncompleteReadError(self._data[old_cursor:], count)

    def reset(self) -> None:
        self._cursor = 0


def create_memory_text_reader(data: str, encoding: str = "utf-8") -> AsyncTextReader:
    return AsyncTextReader(MemoryBytesReader(data.encode(encoding)), encoding)


class MemoryBytesWriter(AsyncBytesWriter):
    _items: List[bytes]

    def __init__(self) -> None:
        self._items = []

    async def write(self, data: bytes) -> None:
        self._items.append(data)

    async def close(self) -> None:
        pass

    def items(self) -> List[bytes]:
        return self._items


def create_memory_text_writer(encoding: str = "utf-8") -> AsyncTextWriter:
    return AsyncTextWriter(MemoryBytesWriter())


class StreamBytesReader(AsyncBytesReader):
    """
    An implementation of `AsyncBytesReader` based on `asyncio.StreamReader`.
    """

    stream_reader: asyncio.StreamReader

    def __init__(self, stream_reader: asyncio.StreamReader) -> None:
        self.stream_reader = stream_reader

    async def read_until(self, separator: bytes = b"\n") -> bytes:
        # StreamReader.readuntil() may raise when its internal buffer cannot hold
        # all the input data. We need to explicitly handle the raised exceptions
        # by "parking" all partial-read results in memory.
        chunks = []
        while True:
            try:
                chunk = await self.stream_reader.readuntil(separator)
                chunks.append(chunk)
                break
            except asyncio.LimitOverrunError as error:
                chunk = await self.stream_reader.readexactly(error.consumed)
                chunks.append(chunk)
        return b"".join(chunks)

    async def read_exactly(self, count: int) -> bytes:
        return await self.stream_reader.readexactly(count)


class StreamBytesWriter(AsyncBytesWriter):
    """
    An implementation of `AsyncBytesWriter` based on `asyncio.StreamWriter`.
    """

    stream_writer: asyncio.StreamWriter

    def __init__(self, stream_writer: asyncio.StreamWriter) -> None:
        self.stream_writer = stream_writer

    async def write(self, data: bytes) -> None:
        self.stream_writer.write(data)
        await self.stream_writer.drain()

    async def close(self) -> None:
        self.stream_writer.close()
        await self._stream_writer_wait_closed()

    async def _stream_writer_wait_closed(self) -> None:
        """
        StreamWriter does not have a `wait_closed` method prior to python
        3.7. For 3.6 compatibility we have to hack it with an async busy
        loop that waits
        - first for the transport to be aware that it is closing
        - then for the socket to become unmapped

        This approach is inspired by the solution in qemu.aqmp.util.
        """
        if sys.version_info >= (3, 7):
            return await self.stream_writer.wait_closed()

        while not self.stream_writer.transport.is_closing():
            await asyncio.sleep(0)

        transport_socket: sys.IO = self.stream_writer.transport.get_extra_info("socket")
        if transport_socket is not None:
            while transport_socket.fileno() != -1:
                await asyncio.sleep(0)


@contextlib.asynccontextmanager
async def _connect_async_bytes(
    socket_path: Path, buffer_size: Optional[int] = None
) -> AsyncIterator[Tuple[AsyncBytesReader, AsyncBytesWriter]]:
    """
    Connect to the socket at given path. Once connected, create an input and
    an output stream from the socket. Both the input stream and the output
    stream are in raw binary mode. The API is intended to be used like this:

    ```
    async with connect(socket_path) as (input_stream, output_stream):
        # Read from input_stream and write into output_stream here
        ...
    ```

    The optional `buffer_size` argument determines the size of the input buffer
    used by the returned reader instance. If not specified, a default value of
    64kb will be used.

    Socket creation, connection, and closure will be automatically handled
    inside this context manager. If any of the socket operations fail, raise
    `ConnectionFailure`.
    """
    writer: Optional[AsyncBytesWriter] = None
    try:
        limit = buffer_size if buffer_size is not None else 2**16
        stream_reader, stream_writer = await asyncio.open_unix_connection(
            str(socket_path), limit=limit
        )
        reader = StreamBytesReader(stream_reader)
        writer = StreamBytesWriter(stream_writer)
        yield reader, writer
    except OSError as error:
        raise ConnectionFailure() from error
    finally:
        if writer is not None:
            await writer.close()


@contextlib.asynccontextmanager
async def connect_async(
    socket_path: Path, buffer_size: Optional[int] = None
) -> AsyncIterator[Tuple[AsyncTextReader, AsyncTextWriter]]:
    """
    This is a line-oriented higher-level API than `connect`. It can be used
    when the caller does not want to deal with the complexity of binary I/O.

    The behavior is the same as `connect`, except the streams that are created
    operates in text mode. Read/write APIs of the streams uses UTF-8 encoded
    `str` instead of `bytes`.
    """
    async with _connect_async_bytes(socket_path, buffer_size) as (
        bytes_reader,
        bytes_writer,
    ):
        yield (
            AsyncTextReader(bytes_reader, encoding="utf-8"),
            AsyncTextWriter(bytes_writer, encoding="utf-8"),
        )


async def create_async_stdin_stdout() -> Tuple[AsyncTextReader, AsyncTextWriter]:
    """
    By default, `sys.stdin` and `sys.stdout` are synchronous channels: reading
    from `sys.stdin` or writing to `sys.stdout` will block until the read/write
    succeed, which is very different from the async socket channels created via
    `connect_async`.

    This function creates wrappers around `sys.stdin` and `sys.stdout` and makes
    them behave in the same way as other async socket channels. This makes it
    easier to write low-level-I/O-agonstic code, where the high-level logic does
    not need to worry about whether the underlying async I/O channel comes from
    sockets or from stdin/stdout.
    """
    loop = asyncio.get_event_loop()
    stream_reader = asyncio.StreamReader(loop=loop)
    await loop.connect_read_pipe(
        lambda: asyncio.StreamReaderProtocol(stream_reader), sys.stdin
    )
    w_transport, w_protocol = await loop.connect_write_pipe(
        asyncio.streams.FlowControlMixin, sys.stdout
    )
    stream_writer = asyncio.StreamWriter(w_transport, w_protocol, stream_reader, loop)
    return (
        AsyncTextReader(StreamBytesReader(stream_reader)),
        AsyncTextWriter(StreamBytesWriter(stream_writer)),
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
from pathlib import Path
from typing import Optional

from pyre_extensions import override
from typing_extensions import Final

from ..configuration import Configuration
from ..filesystem import find_files, LocalMode, path_exists
from ..repository import Repository
from .command import CommandArguments, ErrorSource, ErrorSuppressingCommand

LOG: logging.Logger = logging.getLogger(__name__)


class ExpandTargetCoverage(ErrorSuppressingCommand):
    def __init__(
        self,
        command_arguments: CommandArguments,
        *,
        repository: Repository,
        local_configuration: Optional[str],
        fixme_threshold: bool,
        target_prefix: str,
    ) -> None:
        super().__init__(command_arguments, repository)
        self._local_configuration: Final[Optional[str]] = local_configuration
        self._fixme_threshold: bool = fixme_threshold
        self._target_prefix: str = target_prefix

    @staticmethod
    def from_arguments(
        arguments: argparse.Namespace, repository: Repository
    ) -> "ExpandTargetCoverage":
        command_arguments = CommandArguments.from_arguments(arguments)
        return ExpandTargetCoverage(
            command_arguments,
            repository=repository,
            local_configuration=arguments.local_configuration,
            fixme_threshold=arguments.fixme_threshold,
            target_prefix=arguments.target_prefix,
        )

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(ExpandTargetCoverage, cls).add_arguments(parser)
        parser.set_defaults(command=cls.from_arguments)
        parser.add_argument(
            "-l",
            "--local-configuration",
            type=path_exists,
            help="Path to project root with local configuration",
        )
        parser.add_argument(
            "--fixme-threshold",
            type=int,
            help="Ignore all errors in a file if fixme count exceeds threshold.",
        )
        parser.add_argument(
            "--target-prefix",
            type=str,
            help="The prefix to include in the expanded target.",
        )

    @override
    def run(self) -> None:
        local_root = self._local_configuration
        local_root = Path(local_root) if local_root else Path.cwd()

        # Do not change if configurations exist below given root
        existing_configurations = find_files(local_root, ".pyre_configuration.local")
        if existing_configurations and not existing_configurations == [
            str(local_root / ".pyre_configuration.local")
        ]:
            LOG.warning(
                "Cannot expand targets because nested configurations exist:\n%s",
                "\n".join(existing_configurations),
            )
            return

        # Expand coverage
        local_configuration = Configuration.find_local_configuration(local_root)
        if not local_configuration:
            LOG.warning("Could not find a local configuration to codemod.")
            return
        LOG.info("Expanding typecheck targets in `%s`", local_configuration)
        configuration = Configuration(local_configuration)
        existing_targets = configuration.targets
        glob_target = "{}//{}/...".format(self._target_prefix, str(local_root))
        if existing_targets == [glob_target]:
            LOG.info("Configuration is already fully expanded.")
            return
        configuration.add_targets([glob_target])
        configuration.deduplicate_targets()
        configuration.write()

        # Suppress errors
        self._get_and_suppress_errors(
            configuration,
            error_source=ErrorSource.GENERATE,
            fixme_threshold=self._fixme_threshold,
            fixme_threshold_fallback_mode=LocalMode.IGNORE,
        )

        self._repository.commit_changes(
            commit=(not self._no_commit),
            title=f"Expand target type coverage in {local_root}",
            summary="Expanding type coverage of targets in configuration.",
            set_dependencies=False,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains specialized exceptions for when we encounter invalid
configurations.
"""


class InvalidConfiguration(Exception):
    def __init__(self, message: str) -> None:
        self.message = f"Invalid configuration: {message}"
        super().__init__(self.message)


class InvalidPythonVersion(InvalidConfiguration):
    def __init__(self, message: str) -> None:
        super().__init__(message)
